from abc import ABC, abstractmethod
class User:
    """
    Die Klasse User dient zur Speicherung von
    Usern in einem Online-Shop.
    """
    def __init__(self, idNeu, nachnameNeu, vornameNeu, alter):
        """
        Konstruktor fuer Objekte der Klasse User

        :param idNeu: Identifikationsnummer
        :param nachnameNeu: Nachname
        :param vornameNeu: Vorname
        """
        self.id = idNeu
        self.nachname = nachnameNeu
        self.vorname = vornameNeu
        self.alter = alter

    def InformationAusgeben(self):
        """
        Gibt Information ueber das Datenelement zu Kontrollzwecken
        auf das Terminalfenster aus.
        """
        print(f"Der User {self.vorname} {self.nachname} hat die Identifikationsnummer {self.id} und ist {self.alter} Jahre alt.")

    def IstGleich(self, dVergleich):
        """
        Vergleicht zwei Datenelemente auf Gleichheit.
        :param dVergleich: Datenelement mit dem das Objekt verglichen wird.
        :return: True, wenn die beiden Datenelemente gleichen Schluessel haben.
        """
        return self.id == dVergleich.IdGeben()

    def IstGroesserAls(self, dVergleich):
        """
        Vergleicht zwei Datenelemente bezueglich der Ordnungsrelation.
        :param dVergleich: Datenelement mit dem das Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen groesseren Schluessel hat, als das angegebene Vergleichselement.
        """
        return self.id > dVergleich.IdGeben()

    def SchluesselIstGleich(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, auf Gleichheit.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn die beiden Schluessel gleich sind.
        """
        return self.id == vergleichsSchluessel

    def SchluesselIstGroesserAls(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, bezueglich der Ordnungsrelation.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen groesseren Schluessel hat, als das angegebene Vergleichselement.
        """
        return self.id > vergleichsSchluessel

    def IdGeben(self):
        """
        Geben-Methode zum Attribut id
        :return: Identifikationsnummer
        """
        return self.id

    def SchluesselGeben(self):
        """
        Gibt den Schluessel als Zeichenkette zurueck.
        :return: Vorname
        """
        return str(self.id)

    def NachnameGeben(self):
        """
        Geben-Methode zum Attribut nachname
        :return: Nachname
        """
        return self.nachname

    def VornameGeben(self):
        """
        Geben-Methode zum Attribut vorname
        :return: Vorname
        """
        return self.vorname
    
    def AlterGeben(self):
        """
        Geben-Methode zum Attribut alter
        :return: Alter
        """
        return self.alter


class Baumelement(ABC):
    """
    Abstrakte Klasse Baumelement zur Realisierung einer Baumstruktur mit Hilfe des Entwurfsmusters
    Kompositum: Die Klasse legt ausschliesslich die Schnittstelle fuer die Klassen
    Knoten und Abschluss in Form von Methoden fest.
    """
    def __init__(self):
        """
        Der Konstruktor ist nur der Vollstaendigkeit halber angegeben.
        """
        pass  # nichts zu tun

    @abstractmethod
    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement, falls sein Schluessel noch nicht vorhanden ist,  
        in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binaerbaum 
        bleibt erhalten.
        :param datenNeu: das Datenelement des einzufuegenden Knoten
        :return: (neues) Nachfolgeelement des Aufrufers, None falls es kein Datenelement mit dem eingegebenen Schluessel gibt
        """
        pass

    @abstractmethod
    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel identifiziert wird
        und gibt dieses aus.
        :param suchSchluessel: Schluessel nach dem gesucht wird als Zeichenkette 
        :return: gesuchtes Datenelement bzw. None, falls die Suche erfolglos ist
        """
        pass

    @abstractmethod
    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist. 
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: True, im Erfolgsfall, False sonst.
        """
        pass

    @abstractmethod
    def HoeheGeben(self, aktTiefe):
        """
        Berechnet die Hoehe des Baums und gibt diesen Wert zurueck
        :param aktTiefe: Tiefe des aktuellen Knotens
        :return: die Hoehe = maximale Tiefe (der bisher besuchten Knoten)
        """
        pass

    @abstractmethod
    def TiefeGeben(self, suchSchluessel, tiefe):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette
        :param tiefe: die Tiefe des aktuellen Knotens
        :return: Tiefe als positive Zahl, bzw. -1 falls der suchSchluessel nicht vorhanden ist.
        """
        pass

    @abstractmethod
    def InorderAusgeben(self):
        """
        Gibt die Datenelemente des Baumes in Inorder-Reihenfolge aus.
        """
        pass

    @abstractmethod
    def PreorderAusgeben(self):
        """
        Gibt die Datenelemente des Baumes in Preorder-Reihenfolge aus.
        """
        pass


class Abschluss(Baumelement):
    """
    Klasse Abschluss als Bestandteil eines geordneten Binaerbaums.
    (Implementierung mit dem Entwurfsmuster Kompositum)
    """
    def __init__(self):
        super().__init__()

    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement in die bestehende Datenstruktur ein, indem es einen neuen Knoten 
        (mit zwei Abschluss-Objekten als Nachfolger) erzeugt und eine Referenz auf diesen Knoten
        zurueckgibt. Das Abschluss-Objekt, dass diese Methode ausfuehrt, wird dabei aus dem Baum entfernt.
        :param datenNeu: einzufuegendes Datenelement 
        :return: Referenz auf den neu erzeugten Knoten
        """
        return Knoten(datenNeu)  # Hier waere die Implementierung von Knoten notwendig

    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel identifiziert wird
        und gibt dieses aus.
        :param suchSchluessel: Schluessel nach dem gesucht wird als Zeichenkette 
        :return: None, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return None

    def HoeheGeben(self, aktTiefe):
        """
        Berechnet die Hoehe des Baums und gibt diesen Wert zurueck
        :param aktTiefe: Tiefe des aktuellen Knotens
        :return: die uebergebene Hoehe -1, da hier beim Abschluss das Ende eines Astes erreicht ist und 
        die Abschlusselemente bei der Hoehenberechnung nicht einbezogen werden.
        """
        return aktTiefe - 1

    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist. 
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: False, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return False

    def TiefeGeben(self, suchSchluessel, tiefe):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette
        :param tiefe: aktuelle Tiefe im Baum
        :return: -1 da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return -1
    
    def InorderAusgeben(self):
        pass

    def PreorderAusgeben(self):
        pass


class Knoten(Baumelement):
    """
    Klasse Knoten als Bestandteil eines geordneten Binaerbaums.
    (Implementierung mit Entwurfsmuster Kompositum)
    """
    def __init__(self, datenNeu, linkerNachfolgerNeu=None, rechterNachfolgerNeu=None):
        """
        Konstruktor fuer Objekte der Klasse Knoten ohne linken und rechten Nachfolger, d.h. es werden
        zwei Abschluss-Objekte erzeugt und den entsprechenden Referenzattributen zugewiesen.
        :param datenNeu: Referenz auf ein Objekt der Klasse Woerterbucheintrag, das vom Knoten verwaltet wird.
        :param linkerNachfolgerNeu: der linke Nachfolger des Knotens (default ist ein neuer Abschluss)
        :param rechterNachfolgerNeu: der rechte Nachfolger des Knotens (default ist ein neuer Abschluss)
        """
        super().__init__()
        self.daten = datenNeu
        self.linkerNachfolger = linkerNachfolgerNeu if linkerNachfolgerNeu else Abschluss()
        self.rechterNachfolger = rechterNachfolgerNeu if rechterNachfolgerNeu else Abschluss()

    def Einfuegen(self, datenNeu):
        if self.daten.IstGleich(datenNeu):
            print("Die Daten existieren schon!")
        elif self.daten.IstGroesserAls(datenNeu):
            self.linkerNachfolger = self.linkerNachfolger.Einfuegen(datenNeu)
        else:
            self.rechterNachfolger = self.rechterNachfolger.Einfuegen(datenNeu)
        return self

    def Suchen(self, suchSchluessel):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return self.daten
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.Suchen(suchSchluessel)
        else:
            return self.rechterNachfolger.Suchen(suchSchluessel)

    def IstVorhanden(self, suchSchluessel):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return True
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.IstVorhanden(suchSchluessel)
        else:
            return self.rechterNachfolger.IstVorhanden(suchSchluessel)

    def HoeheGeben(self, aktTiefe):
        linkeHoehe = self.linkerNachfolger.HoeheGeben(aktTiefe + 1)
        rechteHoehe = self.rechterNachfolger.HoeheGeben(aktTiefe + 1)
        return max(linkeHoehe, rechteHoehe)

    def TiefeGeben(self, suchSchluessel, tiefe):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return tiefe
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.TiefeGeben(suchSchluessel, tiefe + 1)
        else:
            return self.rechterNachfolger.TiefeGeben(suchSchluessel, tiefe + 1)
        
    def InorderAusgeben(self):
        self.linkerNachfolger.InorderAusgeben()
        self.daten.InformationAusgeben()
        self.rechterNachfolger.InorderAusgeben()

    def PreorderAusgeben(self):
        self.daten.InformationAusgeben()
        self.linkerNachfolger.PreorderAusgeben()
        self.rechterNachfolger.PreorderAusgeben()


class BinBaum:
    """
    Die Klasse Binbaum ist die Grundstruktur eines geordneten Binaerbaums.
    Konkret wird hier der Binaerbaum genutzt um ein einfaches Woerterbuch umzusetzen,
    deshalb sind die Daten von der Klasse Woerterbucheintrag.
    Die Methoden die ein Objekt dieser Klasse anbietet, leiten den Aufruf an den
    Wurzelknoten weiter, wenn der Baum nicht leer ist.
    (Implementierung mit Entwurfsmuster Kompositum)
    """
    def __init__(self):
        """
        Konstruktor fuer Objekte der Klasse Binbaum
        Ein leerer Baum wird erzeugt: Er besteht nur aus einem Abschluss-Objekt.
        """
        self.wurzel = Abschluss()

    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement sortiert in den geordneten Binaerbaum ein,
        d.h. die Eigenschaft, dass der Baum geordnet ist bleibt bei jedem
        Einfuegevorgang erhalten.
        :param datenNeu: neues Datenelement
        """
        self.wurzel = self.wurzel.Einfuegen(datenNeu)

    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel
        identifiziert wird, und gibt es aus.
        :param suchSchluessel: Schluessel, nach dem gesucht wird
        :return: gesuchtes Datenelement bzw. None, falls die Suche erfolglos ist
        """
        return self.wurzel.Suchen(suchSchluessel)

    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: True, im Erfolgsfall, False sonst.
        """
        return self.wurzel.IstVorhanden(suchSchluessel)

    def HoeheGeben(self):
        """
        Berechnet die Hoehe des Baums und gibt diesen Wert zurueck
        :return: Hoehe des Baums 
        """
        return self.wurzel.HoeheGeben(0)

    def TiefeGeben(self, suchSchluessel):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: Tiefe als positive Zahl, bzw. -1 falls der suchSchluessel nicht vorhanden ist.
        """
        return self.wurzel.TiefeGeben(suchSchluessel, 0)
    
    def InorderAusgeben(self):
        """
        Gibt die Datenelemente des Baumes in Inorder-Reihenfolge aus.
        """
        return self.wurzel.InorderAusgeben()
    
    def PreorderAusgeben(self):
        """
        Gibt die Datenelemente des Baumes in Preorder-Reihenfolge aus.
        """
        return self.wurzel.PreorderAusgeben()

if __name__ == "__main__":
    # Erzeugen von User-Instanzen
    user1 = User(5781, "Geber", "Ann", 17)
    user2 = User(4221, "Tross", "Albert", 30)
    user3 = User(6345, "Fall", "Klara", 50)
    user4 = User(5799, "Ter", "Lee", 14)
    user5 = User(6300, "Sehr", "Kai", 74)
    user6 = User(5888, "Pfahl", "Marta", 61)

    # Erzeugen der BinBaum-Instanz
    binBaum1 = BinBaum()

    # Einfuegen von User-Instanzen in den BinBaum
    binBaum1.Einfuegen(user1)
    binBaum1.Einfuegen(user2)
    binBaum1.Einfuegen(user3)
    binBaum1.Einfuegen(user4)

    print("##########")
    print("Preorder-Ausgabe:")
    binBaum1.PreorderAusgeben()
    
    print("##########")
    print("Inorder-Ausgabe:")
    binBaum1.InorderAusgeben()