from graphics_and_games_klassen import Kreis, Text, Rechteck, Dreieck
import math


class KnotenSymbol():
    """
    Verwaltet ein Knotensymbol in einem Graph
    """

    
    def __init__(self, x, y, radius, farbe, bezeichner):
        """
        Legt das Symbol an und besetzt die Attribute.
        :param x: x-Koordinate des Mittelpunkts
        :param y: y-Koordinate des Mittelpunkts
        :param radius: Radius des Symbols
        :param farbe: Farbe des Symbols
        :param bezeichner: Bezeichner des Symbols
        """
        self.x = x
        self.y = y
        self.r = radius
        self.farbe = farbe
        self.bezeichner = bezeichner
        self.aussen = Kreis()
        self.aussen.FarbeSetzen("schwarz")
        self.innen = Kreis()
        self.text = Text()
        self.text.TextGroesseSetzen(18)
        self.DarstellungAktualisieren()

    
    def DarstellungAktualisieren(self):
        """
        Aktualisiert das Symbol
        """
        self.aussen.PositionSetzen(self.x, self.y)
        self.aussen.RadiusSetzen(self.r)
        self.innen.PositionSetzen(self.x, self.y)
        self.innen.RadiusSetzen(self.r - 2)
        self.innen.FarbeSetzen(self.farbe)
        self.text.PositionSetzen(int(self.x - len(self.bezeichner) / 2 * 10 - len(self.bezeichner) % 2 * 5), self.y + 10)
        self.text.TextSetzen(self.bezeichner)

    def FarbeSetzen(self, f):
        """
        Setzt die Farbe der Darstellung
        """
        self.farbe = f
        self.DarstellungAktualisieren()

    def PositionSetzen(self, x, y):
        """
        Setzt die Position der Darstellung
        """
        self.x = x
        self.y = y
        self.DarstellungAktualisieren()

    def BezeichnerSetzen(self, bezeichner):
        """
        Setzt die Farbe der Darstellung
        """
        self.bezeichner = bezeichner
        self.DarstellungAktualisieren()

    def XGeben(self):
        """
        Meldet die x-Koordinate des Symbols zurück.
        """
        return self.x

    def YGeben(self):
        """
        Meldet die y-Koordinate des Symbols zurück.
        """
        return self.y

    def BezeichnerGeben(self):
        """
        Meldet den Bezeichner des Symbols zurück.
        """
        return self.bezeichner

    def Entfernen(self):
        """
        Entfernt das Knotensymbol aus der Anzeige
        """
        self.aussen.Entfernen()
        self.innen.Entfernen()
        self.text.Entfernen()


class KantenSymbol():
    """
    Verwaltet ein Kantensymbol in einem Graph"""

    def __init__(self, start, ende, gerichtet, gewicht, breite, farbe):
        """
        Legt das Symbol an und besetzt die Attribute.
        :param start: Startknoten
        :param ende: Zielknoten
        :param gerichtet: True, wenn die Kante gerichtet ist
        :param gewicht: Gewicht der Kante
        :param breite: Breite der Kante
        :param farbe: Farbe der Kante
        """
        self.von = start
        self.nach = ende
        self.farbe = farbe
        self.breite = breite
        self.gerichtet = gerichtet
        self.gewicht = gewicht
        self.aussen = Rechteck()
        self.innen = Rechteck()
        self.pfeil = Dreieck()
        self.pfeil.FarbeSetzen("schwarz")
        self.text = Text()
        self.text.SichtbarkeitSetzen(gewicht != "")
        self.text.TextSetzen(gewicht)
        self.DarstellungAktualisieren()

    def DarstellungAktualisieren(self):
        """
        Aktualisiert das Symbol
        """
        x1 = self.von.XGeben()
        y1 = self.von.YGeben()
        x2 = self.nach.XGeben()
        y2 = self.nach.YGeben()
        laenge = int(math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)))
        xm = int((x1 + x2) / 2.0)
        ym = int((y1 + y2) / 2.0)
        x = int(xm - laenge / 2)
        y = int(ym - self.breite / 2)
        self.aussen.PositionSetzen(x, y)
        self.aussen.GroesseSetzen(laenge, self.breite)
        if (self.breite < 4):
            groesse = self.breite * 4
            self.innen.SichtbarkeitSetzen(False)
            self.aussen.FarbeSetzen(self.farbe)
            self.pfeil.SichtbarkeitSetzen(self.gerichtet)
            self.pfeil.PositionSetzen(xm, int(ym - groesse / 2))
            self.pfeil.GroesseSetzen(groesse, groesse)
            self.pfeil.WinkelSetzen(270)
        else:
            delta = 2 if self.breite <= 8 else 4
            self.innen.SichtbarkeitSetzen(True)
            self.innen.PositionSetzen(x, int(y + delta / 2))
            self.innen.GroesseSetzen(laenge, self.breite - delta)
            self.innen.FarbeSetzen(self.farbe)
            self.aussen.FarbeSetzen("schwarz")
            self.pfeil.SichtbarkeitSetzen(self.gerichtet)
            self.pfeil.PositionSetzen(xm, int(ym - self.breite / 2))
            self.pfeil.GroesseSetzen(self.breite - delta, self.breite)
            self.pfeil.WinkelSetzen(270)

        winkel = 0.0
        if x1 == x2:
            if y2 > y1:
                winkel = -90
            else:
                winkel = 90
        elif x1 < x2:
            winkel = -math.atan((y2 - y1) / (x2 - x1)) / math.pi * 180.0
        else:
            winkel = 180 - math.atan((y2 - y1) / (x2 - x1)) / math.pi * 180.0
        self.pfeil.WinkelSetzen(int(winkel + 270))
        self.innen.WinkelSetzen(int(winkel))
        self.aussen.WinkelSetzen(int(winkel))
        self.pfeil.GanzNachHintenBringen()
        self.innen.GanzNachHintenBringen()
        self.aussen.GanzNachHintenBringen()
        if ((-90 <= winkel) and (winkel <= 0)):
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm + int(delta * math.cos(w)), ym - int(delta * math.sin(w)))
        elif ((0 < winkel) and (winkel <= 90)):
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm - int(delta * math.cos(w)), ym + int(delta * math.sin(w)) + 7)
        elif ((90 < winkel) and (winkel <= 180)):
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm - int(delta * math.cos(w)), ym + int(delta * math.sin(w)))
        else:
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm + int(delta * math.cos(w)), ym - int(delta * math.sin(w)) + 10)
    
    def Entfernen(self):
        """
        Entfernt das Kantensymbol aus der Anzeige
        """
        self.aussen.Entfernen()
        self.innen.Entfernen()
        self.pfeil.Entfernen()
        self.text.Entfernen()

    def FarbeSetzen(self, f):
        """
        Setzt die Farbe der Darstellung
        """
        self.farbe = f
        self.DarstellungAktualisieren()

    def StartsymbolGeben(self):
        """
        Meldet den Startknoten
        """
        return self.von

    def ZielsymbolGeben(self):
        """
        Meldet den Zielknoten
        """
        return self.nach