from abc import ABC, abstractmethod

class Woerterbucheintrag:
    """
    Die Klasse Woerterbucheintrag dient zur Speicherung von
    Woerterbucheintraegen fuer ein Englisch-Deutsches
    Woerterbuch. Dabei kann jedem englischen Wort mehrere
    deutsche Bedeutungen zugeordnet werden, wobei alle Bedeutungen in
    einer einzigen Zeichenkette Bedeutung gespeichert werden.
    """
    def __init__(self, wortNeu, bedeutungNeu):
        """
        Konstruktor fuer Objekte der Klasse Woerterbucheintrag
        :param wortNeu: das englische Wort.
        :param bedeutungNeu: die deutsche(n) Bedeutung(en).
        """
        self.wort = wortNeu
        self.bedeutung = bedeutungNeu

    def InformationAusgeben(self):
        """
        Gibt Information ueber das Datenelement zu Kontrollzwecken
        auf das Terminalfenster aus.
        """
        print(f"{self.wort}: {self.bedeutung}")

    def IstGleich(self, dvergleich):
        """
        Vergleicht zwei Datenelemente auf Gleichheit.
        :param dvergleich: Datenelement mit dem das Objekt verglichen wird.
        :return: True, wenn die beiden Datenelemente gleichen Schluessel haben.
        """
        return self.wort == dvergleich.WortGeben()

    def IstGroesserAls(self, dvergleich):
        """
        Vergleicht zwei Datenelemente bezueglich der Ordnungsrelation.
        :param vergleichselement: Datenelement mit dem das Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen groesseren Schluessel hat, als das angegebene Vergleichselement.
        """
        return self.wort > dvergleich.WortGeben()

    def SchluesselIstGleich(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, auf Gleichheit.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn die beiden Schluessel gleich sind.
        """
        return self.wort == vergleichsSchluessel

    def SchluesselIstGroesserAls(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, bezueglich der Ordnungsrelation.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen groesseren Schluessel hat, als das angegebene Vergleichselement.
        """
        return self.wort > vergleichsSchluessel

    def SchluesselIstKleinerAls(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, bezueglich der Ordnungsrelation.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen kleineren Schluessel 
        hat, als das angegebene Vergleichselement.
        """
        return self.wort < vergleichsSchluessel

    def WortGeben(self):
        """
        Geben-Methode zum Attribut wort
        :return: das englische Wort
        """
        return self.wort

    def BedeutungGeben(self):
        """
        Geben-Methode zum Attribut bedeutung
        :return: bedeutung: die deutsche(n) Bedeutung(en) als eine Zeichenkette
        """
        return self.bedeutung

    def BedeutungSetzen(self, bedeutungNeu):
        """
        Setzen-Methode zum Attribut bedeutung
        :param bedeutungNeu: neue Bedeutung fuer das englische Wort
        """
        self.bedeutung = bedeutungNeu


class Baumelement(ABC):
    """
    Abstrakte Klasse Baumelement zur Realisierung einer Baumstruktur mit Hilfe des Entwurfsmusters
    Kompositum: Die Klasse legt ausschliesslich die Schnittstelle fuer die Klassen
    Knoten und Abschluss in Form von Methoden fest.
    """
    def __init__(self):
        """
        Der Konstruktor ist nur der Vollstaendigkeit halber angegeben.
        """
        pass  # nichts zu tun

    @abstractmethod
    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement, falls sein Schluessel noch nicht vorhanden ist,  
        in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binaerbaum 
        bleibt erhalten.
        :param datenNeu: das Datenelement des einzufuegenden Knoten
        :return: (neues) Nachfolgeelement des Aufrufers, None falls es kein Datenelement mit dem eingegebenen Schluessel gibt
        """
        pass

    @abstractmethod
    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel identifiziert wird
        und gibt dieses aus.
        :param suchSchluessel: Schluessel nach dem gesucht wird als Zeichenkette 
        :return: gesuchtes Datenelement bzw. None, falls die Suche erfolglos ist
        """
        pass

    @abstractmethod
    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist. 
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: True, im Erfolgsfall, False sonst.
        """
        pass

    @abstractmethod
    def HoeheGeben(self, aktTiefe):
        """
        Berechnet die Hoehe des Baums und gibt diesen Wert zurueck
        :param aktTiefe: Tiefe des aktuellen Knotens
        :return: die Hoehe = maximale Tiefe (der bisher besuchten Knoten)
        """
        pass

    @abstractmethod
    def TiefeGeben(self, suchSchluessel, tiefe):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette
        :param tiefe: die Tiefe des aktuellen Knotens
        :return: Tiefe als positive Zahl, bzw. -1 falls der suchSchluessel nicht vorhanden ist.
        """
        pass


class Abschluss(Baumelement):
    """
    Klasse Abschluss als Bestandteil eines geordneten Binaerbaums.
    (Implementierung mit dem Entwurfsmuster Kompositum)
    """
    def __init__(self):
        super().__init__()

    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement in die bestehende Datenstruktur ein, indem es einen neuen Knoten 
        (mit zwei Abschluss-Objekten als Nachfolger) erzeugt und eine Referenz auf diesen Knoten
        zurueckgibt. Das Abschluss-Objekt, dass diese Methode ausfuehrt, wird dabei aus dem Baum entfernt.
        :param datenNeu: einzufuegendes Datenelement 
        :return: Referenz auf den neu erzeugten Knoten
        """
        return Knoten(datenNeu)  # Hier waere die Implementierung von Knoten notwendig

    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel identifiziert wird
        und gibt dieses aus.
        :param suchSchluessel: Schluessel nach dem gesucht wird als Zeichenkette 
        :return: None, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return None

    def HoeheGeben(self, aktTiefe):
        """
        Berechnet die Hoehe des Baums und gibt diesen Wert zurueck
        :param aktTiefe: Tiefe des aktuellen Knotens
        :return: die uebergebene Hoehe -1, da hier beim Abschluss das Ende eines Astes erreicht ist und 
        die Abschlusselemente bei der Hoehenberechnung nicht einbezogen werden.
        """
        return aktTiefe - 1

    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist. 
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: False, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return False

    def TiefeGeben(self, suchSchluessel, tiefe):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette
        :param tiefe: aktuelle Tiefe im Baum
        :return: -1 da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return -1

class Knoten(Baumelement):
    """
    Klasse Knoten als Bestandteil eines geordneten Binaerbaums.
    (Implementierung mit Entwurfsmuster Kompositum)
    """
    def __init__(self, datenNeu, linkerNachfolgerNeu=None, rechterNachfolgerNeu=None):
        """
        Konstruktor fuer Objekte der Klasse Knoten ohne linken und rechten Nachfolger, d.h. es werden
        zwei Abschluss-Objekte erzeugt und den entsprechenden Referenzattributen zugewiesen.
        :param datenNeu: Referenz auf ein Objekt der Klasse Woerterbucheintrag, das vom Knoten verwaltet wird.
        :param linkerNachfolgerNeu: der linke Nachfolger des Knotens (default ist ein neuer Abschluss)
        :param rechterNachfolgerNeu: der rechte Nachfolger des Knotens (default ist ein neuer Abschluss)
        """
        super().__init__()
        self.daten = datenNeu
        self.linkerNachfolger = linkerNachfolgerNeu if linkerNachfolgerNeu else Abschluss()
        self.rechterNachfolger = rechterNachfolgerNeu if rechterNachfolgerNeu else Abschluss()

    def Einfuegen(self, datenNeu):
        if self.daten.IstGleich(datenNeu):
            print("Die Daten existieren schon!")
        elif self.daten.IstGroesserAls(datenNeu):
            self.linkerNachfolger = self.linkerNachfolger.Einfuegen(datenNeu)
        else:
            self.rechterNachfolger = self.rechterNachfolger.Einfuegen(datenNeu)
        return self

    def Suchen(self, suchSchluessel):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return self.daten
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.Suchen(suchSchluessel)
        else:
            return self.rechterNachfolger.Suchen(suchSchluessel)

    def IstVorhanden(self, suchSchluessel):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return True
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.IstVorhanden(suchSchluessel)
        else:
            return self.rechterNachfolger.IstVorhanden(suchSchluessel)

    def HoeheGeben(self, aktTiefe):
        linkeHoehe = self.linkerNachfolger.HoeheGeben(aktTiefe + 1)
        rechteHoehe = self.rechterNachfolger.HoeheGeben(aktTiefe + 1)
        return max(linkeHoehe, rechteHoehe)

    def TiefeGeben(self, suchSchluessel, tiefe):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return tiefe
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.TiefeGeben(suchSchluessel, tiefe + 1)
        else:
            return self.rechterNachfolger.TiefeGeben(suchSchluessel, tiefe + 1)


class BinBaum:
    """
    Die Klasse Binbaum ist die Grundstruktur eines geordneten Binaerbaums.
    Konkret wird hier der Binaerbaum genutzt um ein einfaches Woerterbuch umzusetzen,
    deshalb sind die Daten von der Klasse Woerterbucheintrag.
    Die Methoden die ein Objekt dieser Klasse anbietet, leiten den Aufruf an den
    Wurzelknoten weiter, wenn der Baum nicht leer ist.
    (Implementierung mit Entwurfsmuster Kompositum)
    """
    def __init__(self):
        """
        Konstruktor fuer Objekte der Klasse Binbaum
        Ein leerer Baum wird erzeugt: Er besteht nur aus einem Abschluss-Objekt.
        """
        self.wurzel = Abschluss()

    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement sortiert in den geordneten Binaerbaum ein,
        d.h. die Eigenschaft, dass der Baum geordnet ist bleibt bei jedem
        Einfuegevorgang erhalten.
        :param datenNeu: neues Datenelement
        """
        self.wurzel = self.wurzel.Einfuegen(datenNeu)

    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel
        identifiziert wird, und gibt es aus.
        :param suchSchluessel: Schluessel, nach dem gesucht wird
        :return: gesuchtes Datenelement bzw. None, falls die Suche erfolglos ist
        """
        return self.wurzel.Suchen(suchSchluessel)

    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: True, im Erfolgsfall, False sonst.
        """
        return self.wurzel.IstVorhanden(suchSchluessel)

    def HoeheGeben(self):
        """
        Berechnet die Hoehe des Baums und gibt diesen Wert zurueck
        :return: Hoehe des Baums 
        """
        return self.wurzel.HoeheGeben(0)

    def TiefeGeben(self, suchSchluessel):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: Tiefe als positive Zahl, bzw. -1 falls der suchSchluessel nicht vorhanden ist.
        """
        return self.wurzel.TiefeGeben(suchSchluessel, 0)


if __name__ == "__main__":
    woerterb1 = Woerterbucheintrag("clip", "Clip, Spange, Klammer")
    woerterb2 = Woerterbucheintrag("cat", "Katze")
    woerterb3 = Woerterbucheintrag("cup", "Tasse, Becher")
    woerterb4 = Woerterbucheintrag("cut", "Schnitt, oeffnung, kuerzen")
    woerterb5 = Woerterbucheintrag("cube", "Wuerfel")
    woerterb6 = Woerterbucheintrag("cave", "Hoehle")
    binBaum1 = BinBaum()
    binBaum1.Einfuegen(woerterb1)
    binBaum1.Einfuegen(woerterb2)
    binBaum1.Einfuegen(woerterb3)
    binBaum1.Einfuegen(woerterb4)